#! /usr/bin/env python

"""A module for sequence motifs."""

import os
import sys
import time
import math
import copy
import distutils.dir_util
import __init__

import numpy

#import lxml.etree as etree
import xml.dom.minidom as minidom


SCHEMA_FILE_NAME = 'motif.xsd'
MUALPHA0 =  0.296 
MUALPHA1 =  0.641
SDALPHA0 =  0.943
SDALPHA1 = -0.111 

#PWM_FILE  = os.path.join(__init__.PWMPATH,'pbm.xml')
PWM_FILE  = os.path.join(__init__.PWMPATH,'transfac.xml')

def information_content(m):
    """
    calculate the information content at every position in a PWM
    returns a list of bits, corresponding to each position 
    """

    y = []
    for i,elem in enumerate(m):
        z = 2
        for j,base in enumerate(elem):
            if base > 0:
                z += base*math.log(base)/math.log(2)

        y += [z]
    return y

def pwm_to_iupac(m):
    """
    Convert a position weight matrix to IUPAC consensus based on 
    at least 0.5 bit of information at informative positions.
    """
  
    # same length for all sequences
    N = 25
    MININFO = 0.5 # minimum number of bits at a position in the PWM
    MINFRAC = 0.1 # fraction of contribution difference of predominant base

    l = [ 'A', 'C', 'G', 'T', 'AG', 'CT', 'CA',\
    'TG', 'TA', 'CG', 'CTG', 'ATG', 'ATC', 'ACG', 'ACGT' ]

#    d = { 'A':'A', 'C':'C', 'G':'G', 'T':'T', 'AG':'R', 'CT':'Y', 'AC':'M',\
#    'GT':'K', 'AT':'W', 'CG':'S', 'CGT':'B', 'AGT':'C', 'ACT':'G', 'ACG':'V', 'ACGT':'.' }

    d = { 'A':'A', 'C':'C', 'G':'G', 'T':'T',\
          'AG':'R', 'CT':'Y', 'AC':'M', 'GT':'K', 'AT':'W', 'CG':'S',\
          'CGT':'.', 'AGT':'.', 'ACT':'.', 'ACG':'.', 'ACGT':'.' }

    base = ['A','C','G','T']

    s = ''

    info = information_content(m)

    for k,elem in enumerate(m):

        if info[k] > MININFO:
            u = sorted( [ (-x,i) for i,x in enumerate(elem) ] )

            #t = 0; z = []
            #for j in range(4):
            #    t += -u[j][0]
            #    z += [base[ u[j][1] ]]
            #    if t > MINFRAC:
            #        break

            if -u[0][0] > -u[1][0] + MINFRAC:
                z = [ base[u[0][1]] ]
            else:
                z = [ base[u[0][1]], base[u[1][1]] ]

            z.sort()
            z = ''.join(z)
            s += d[z]
        else:
            s += d['ACGT']

    if len(s) > N:
        s = s[0:N]
    elif len(s) < N:
        s = s + ''.join( (N-len(s))*[' '] ) 

    return s



class Motif:
    """A class for a sequence motif.
    
    Attributes:
    
    self.id : string, required
        A string containing a unique ID identifying the motif.  
    
    self.status : string, optional
        A string describing the status of the motif, such as "putative", 
        "confirmed", or "obsolete".
    
    self.source : string, optional
        A string describing the source of the motif, such as "Transfac" or 
        "PBM".
    
    self.sourcefile : string, optional
        A string containing the path to the source file of the motif.
    
    self.species : list of strings, optional
        A list of species names, such as ["mouse", "human", "yeast"].
    
    self.entrez : list of integers, optional
        A list of integers representing the Entrez Gene IDs of the transcription 
        factor that bind to the motif, such as [6908, 2099].
    
    self.symbol : string, optional
        The symbol of the transcription factor that binds to the motif, such as 
        'ERa'.
    
    self.synonyms : list of strings, optional
        A list of synonyms for the transcription factor that binds to the motif, 
        such as ['ESR', 'ESR1'].
    
    self.fullname : string, optional
        The full name of the transcription factor that binds to the motif, such 
        as "estrogen receptor alpha".
    
    self.dbd : string, optional
        The full name of the DNA binding domain of the transcription factor that 
        binds to the motif, such as "estrogen receptor region C".
    
    self.pmid : integer, optional
        An integer representing the Pubmed ID of the article in which the motif 
        was published.
    
    self.pssm : NumPy float array (N rows, 4 columns), required
        A NumPy array containing the position specific scoring matrix (PSSM) 
        for the motif.  The array should contain N rows, each corresponding to 
        a single position in the PSSM, and 4 columns, each corresponding to the 
        4 DNA basepairs A, C, G, T, in order.  Each element in the matrix 
        should be a float, and each row of the matrix should sum to 1.0.
    
    self.numseqs : integer, optional
        The number of sequences used to derive the motif's position specific 
        scoring matrix (PSSM).
    
    self.curators : list of strings, optional
        A list of the full names of the curators, such as ["Cliff Meyer", 
        "Xiaole Shirley Liu", "Mayako Michino"].
    
    self.results : dictionary
        A dictionary containing information about the  scores for the 
        motif.  The dictionary should contain the following keys:
          ### ADD THIS LATER ###
          ### (Miyako, the self.seqpos_results attribute isn't saved into the XML file,
               so you won't have to deal withthis. ###"""
    
    ##################
    # initialization #
    ##################
    
    def __init__(self, motif_id=None, motif_pssm=None):
        """Initialize the motif with PSSM m."""
        self.id = motif_id
        self.status = None
        self.source = None
        self.sourcefile = None
        self.species = []
        self.entrez = []
        self.symbol = None
        self.synonyms = []
        self.fullname = None
        self.dbd = None
        self.pmid = None
        if motif_pssm is None:
            self.pssm = None
        else:
            self.pssm = numpy.array(motif_pssm, float)
            if self.pssm.shape[1] != 4:
                raise ValueError, "motif PSSM must have 4 columns"
        self.numseqs = None
        self.curators = []
        self.results = None
        self.antisense = False # reverse complements need to be taken to make tree consistent
    
    def clean_str(self, attribute):
        """Clean up a string attribute."""
        value = getattr(self, attribute)
        if value is None:
            return
        string = str(value)
        if string.isspace():
            setattr(self, attribute, None)
        else:
            setattr(self, attribute, string)
    
    def clean_int(self, attribute):
        """Clean up an integer attribute."""
        value = getattr(self, attribute)
        if value is None:
            return
        try:
            integer = int(value)
            setattr(self, attribute, integer)
        except ValueError:
            print >> sys.stderr, "warning: cannot convert %s value '%s' to an integer" % (attribute, value)
            setattr(self, attribute, None)
    
    def clean_str_list(self, attribute):
        """Clean up a string list attribute."""
        value = getattr(self, attribute)
        if value is None:
            setattr(self, attribute, [])
            return
        if type(value) not in (list, tuple):
            value = [value]
        string_list = []
        for elem in value:
            if elem is None:
                continue
            string = str(elem)
            if not string.isspace():
                string_list.append(string)
        setattr(self, attribute, string_list)
    
    def clean_int_list(self, attribute):
        """Clean up an integer list attribute."""
        value = getattr(self, attribute)
        if attribute is None:
            setattr(self, attribute, [])
            return
        if type(value) not in (list, tuple):
            value = [value]
        integer_list = []
        for elem in value:
            if elem is None:
                continue
            try:
                integer = int(elem)
            except ValueError:
                print >> sys.stderr, "warning: cannot convert %s element '%s' to an integer" % (attribute, elem)
                continue
            integer_list.append(integer)
        setattr(self, attribute, integer_list)
    
    def cleanup(self):
        """Clean up formatting for all motif attributes."""
        for attribute in ('id', 'status', 'source', 'sourcefile', 'symbol', 'fullname', 'dbd'):
            self.clean_str(attribute)
        for attribute in ('pmid', 'numseqs'):
            self.clean_int(attribute)
        for attribute in ('species', 'synonyms', 'curators'):
            self.clean_str_list(attribute)
        for attribute in ('entrez',):
            self.clean_int_list(attribute)
        if self.pssm is not None:
            try:
                self.pssm = numpy.array(self.pssm, float)
            except ValueError:
                print >> sys.stderr, "warning: cannot convert motif PSSM to a NumPy array:"
                print >> sys.stderr, self.pssm
                self.pssm = None
            if self.pssm.shape[1] != 4:
                print >> sys.stderr, "warning: motif PSSM must have 4 columns:"
                self.pssm = None
        
    ##################
    # static methods #
    ##################
 
    @staticmethod
    def isduplicate(motif1, motif2):
        """Return True if two motifs are the same, and False otherwise."""
        return motif1.id == motif2.id
 
 
    #################
    # input methods #
    #################
 
#    def from_xml_file(self, xml_file_name, schema_file_name=SCHEMA_FILE_NAME):
#        """Read list of motifs from XML file."""
#        # parse motif XML
#        tree = etree.parse(xml_file_name)
#        # retrieve root tag
#        root = tree.getroot()
#        if root.tag != 'motif':
#            raise ValueError, "root element of XML should be <motif>"
#        # extract motif data
#        self.from_xml_etree(root)
    
    def from_xml_etree(self, root):
        """Read motif from an XML ElementTree.
        
        Any data that was stored in Motif object will be overwritten."""
        assert root.tag == 'motif'
        # retrieve motif ID
        self.id = root.attrib['id']
        # retrieve status
        self.status = root.find('status')
        if self.status is not None:
            self.status = self.status.text
        # retrieve source
        self.source = root.find('source')
        if self.source is not None:
            self.source = self.source.text
        # retrieve path to source file
        self.sourcefile = root.find('sourcefile')
        if self.sourcefile is not None:
            self.sourcefile = self.sourcefile.text
        # retrieve list of species
        self.species = [species.text for species in root.findall('species')]
        # retrieve Entrez GeneID
        self.entrez = [int(entrez.text) for entrez in root.findall('entrez')]
        # retrieve symbol
        self.symbol = root.find('symbol')
        if self.symbol is not None:
            self.symbol = self.symbol.text
        # retrieve list of synonyms
        self.synonyms = [synonym.text for synonym in root.findall('synonym')]
        # retrieve full name
        self.fullname = root.find('fullname')
        if self.fullname is not None:
            self.fullname = self.fullname.text
        # retrieve DNA binding domain
        self.dbd = root.find('dbd')
        if self.dbd is not None:
            self.dbd = self.dbd.text
        # retrieve PubMed ID
        self.pmid = root.find('pmid')
        if self.pmid is not None:
            self.pmid = int(self.pmid.text)
        # retrieve PSSM
        self.pssm = root.find('pssm')
        if self.pssm is not None:
            pos_elements = self.pssm.findall('pos')
            self.pssm = numpy.empty((len(pos_elements), 4), float)
            for i, pos in enumerate(pos_elements):
                assert int(pos.attrib['num']) == i + 1
                self.pssm[i,:] = [float(base.text) for base in pos]
        # retrieve number of sequences
        self.numseqs = root.find('numseqs')
        if self.numseqs is not None:
            self.numseqs = int(self.numseqs.text)
        # retrieve list of curators
        self.curators = [curator.text for curator in root.findall('curator')]
        # clear results
        self.results = None
    
    ##################
    # output methods #
    ##################
    
    def __str__(self):
        """Generate string representation of motif."""
        motif_string = []
        motif_string.append("Motif ID: %s" % self.id)
        motif_string.append("Status: %s" % self.status)
        motif_string.append("Source: %s" % self.source)
        motif_string.append("Source File: %s" % self.sourcefile)
        if len(self.species) == 0:
            motif_string.append("Species: None")
        else:
            motif_string.append("Species: %s" % ', '.join(self.species))
        if len(self.entrez) == 0:
            motif_string.append("Entrez GeneID: None")
        else:
            motif_string.append("Entrez GeneID: %s" % ', '.join(str(entrez) for entrez in self.entrez))
        motif_string.append("Symbol: %s" % self.symbol)
        if len(self.synonyms) == 0:
            motif_string.append("Synonyms: None")
        else:
            motif_string.append("Synonyms: %s" % ', '.join(self.synonyms))
        motif_string.append("Full Name: %s" % self.fullname)
        motif_string.append("DNA Binding Domain: %s" % self.dbd)
        motif_string.append("PMID: %s" % self.pmid)
        if self.pssm is None:
            motif_string.append("PSSM: None")
        else:
            motif_string.append("PSSM:        A      C      G      T  ")
            for i, pos in enumerate(self.pssm):
                motif_string.append(" |%6d   %4.2f   %4.2f   %4.2f   %4.2f" % (i + 1, pos[0], pos[1], pos[2], pos[3]))
        motif_string.append("Number of Sequences: %s" % self.numseqs)
        if len(self.curators) == 0:
            motif_string.append("Curators: None")
        else:
            motif_string.append("Curators: %s" % ', '.join(self.curators))
        if self.results is None:
            motif_string.append("Results: None")
        else:
            motif_string.append("Analysis Results:")
            for result_key in self.results:
                motif_string.append( '%s ... %.2f' % (result_key,self.results[result_key]) )
            #motif_string.append(" |  Number of Hits ... %d" % self.seqpos_results['numhits'])
            #motif_string.append(" |  Cutoff ........... %.2f" % self.seqpos_results['cutoff'])
            #motif_string.append(" |  Z-score .......... %.2f" % self.seqpos_results['zscore'])
            #motif_string.append(" |  p-value .......... %.2f" % self.seqpos_results['pvalue'])
        motif_string = '\n'.join(motif_string)
        return motif_string

    def to_xml(self, pretty_print=False):
        """Return motif as XML."""
        root = self.to_xml_etree()
        if root is not None:
            motif_as_xml = etree.tostring(root)
            if pretty_print is True:
                motif_as_xml = minidom.parseString(motif_as_xml).toprettyxml(indent="  ")
            return motif_as_xml
    
    def to_xml_etree(self):
        """Return motif as an XML ElementTree."""
        # motif ID and motif PSSM must be defined
        if self.id is None:
            print >> sys.stderr, "motif ID 'self.id' must be defined for conversion to XML"
            return
        if self.pssm is None:
            print >> sys.stderr, "motif PSSM 'self.pssm' must be defined for conversion to XML"
            return
        # root element
        root = etree.Element('motif')
        root.attrib['id'] = str(self.id)
        # status element
        if self.status is not None:
            status_element = etree.SubElement(root, 'status')
            status_element.text = str(self.status)
        # source element
        if self.source is not None:
            source_element = etree.SubElement(root, 'source')
            source_element.text = str(self.source)
        # sourcefile element
        if self.sourcefile is not None:
            sourcefile_element = etree.SubElement(root, 'sourcefile')
            sourcefile_element.text = str(self.sourcefile)
        # species elements
        for species in self.species:
            species_element = etree.SubElement(root, 'species')
            species_element.text = str(species)
        # entrez element
        for entrez in self.entrez:
            entrez_element = etree.SubElement(root, 'entrez')
            entrez_element.text = str(entrez)
        # symbol element
        if self.symbol is not None:
            symbol_element = etree.SubElement(root, 'symbol')
            symbol_element.text = str(self.symbol)
        # synonym elements
        for synonym in self.synonyms:
            synonym_element = etree.SubElement(root, 'synonym')
            synonym_element.text = str(synonym)
        # fullname element
        if self.fullname is not None:
            fullname_element = etree.SubElement(root, 'fullname')
            fullname_element.text = str(self.fullname)
        # dbd element
        if self.dbd is not None:
            dbd_element = etree.SubElement(root, 'dbd')
            dbd_element.text = str(self.dbd)
        # pmid element
        if self.pmid is not None:
            pmid_element = etree.SubElement(root, 'pmid')
            pmid_element.text = str(self.pmid)
        # pssm element
        pssm_element = etree.SubElement(root, 'pssm')
        for i, pos in enumerate(self.pssm):
            pos_element = etree.SubElement(pssm_element, 'pos')
            pos_element.attrib['num'] = str(i + 1)
            A_element = etree.SubElement(pos_element, 'A')
            A_element.text = "%.2f" % pos[0]
            C_element = etree.SubElement(pos_element, 'C')
            C_element.text = "%.2f" % pos[1]
            G_element = etree.SubElement(pos_element, 'G')
            G_element.text = "%.2f" % pos[2]
            T_element = etree.SubElement(pos_element, 'T')
            T_element.text = "%.2f" % pos[3]
        # numseqs element
        if self.numseqs is not None:
            numseqs_element = etree.SubElement(root, 'numseqs')
            numseqs_element.text = str(self.numseqs)
        # curator elements
        for curator in self.curators:
            curator_element = etree.SubElement(root, 'curator')
            curator_element.text = str(curator)
        return root
    

    def logo(self, width, height, logo_dir="img/"):
        """Generate motif logo and return its file name."""
        return 
    
    def to_html(self, img_dir="img/"):
        """Return motif as pretty HTML for display purposes."""
        # generate a motif logo
        logo_file_name = self.logo(192, 120, logo_dir=img_dir)
        # prepare the motif PSSM for display
        motif_pssm = []
        if self.pssm is None:
            motif_pssm.append("PSSM: None")
        else:
            motif_pssm.append("        A      C      G      T  ")
            for i, pos in enumerate(self.pssm_for_tree):
                motif_pssm.append("%3d   %4.2f   %4.2f   %4.2f   %4.2f" % (i + 1, pos[0], pos[1], pos[2], pos[3]))
        # prepare SeqPos results for display
        results = []
        if self.results is None or ( 'numhits' not in self.seqpos_results ):
            results.append("SeqPos Results: None")
        else:
            results.append("SeqPos Results:")
            results.append(" |  Number of Hits ... %d"   % self.results['numhits'])
            results.append(" |  Cutoff ........... %.2f" % self.results['cutoff'])
            results.append(" |  Z-score .......... %.2f" % self.results['zscore'])
            results.append(" |  p-value .......... %.4e" % self.results['pvalue'])
        # load Django template for motifs
        t = loader.get_template('motif.html')
        # load context for Django template
        c = Context({
            'img_dir': img_dir,
            'motif_name': self.id,
            'motif_logo': logo_file_name,
            'motif_pssm': '\n'.join(motif_pssm),
            'results': '\n'.join(results),
        })
        motif_as_html = t.render(c)
        return motif_as_html


class MotifList(list):
    """A class for a list of sequence motifs."""
    
    ##################
    # initialization #
    ##################
    
    def __init__(self, iterable=[]):
        list.__init__(self, iterable)
        self.check_type()
    
    ######################################################
    # inherited methods, supplemented with type checking #
    ######################################################
    
    def __add__(self, another_motif_list):
        new_motif_list = MotifList()
        new_motif_list.extend(self)
        new_motif_list.extend(another_motif_list)
        return new_motif_list
    
    def __setitem__(self, index, value):
        list.__setitem__(self, index, value)
        self.check_type()
    
    def __setslice__(self, index_start, index_end, iterable):
        list.__setslice__(self, index_start, index_end, iterable)
        self.check_type()
    
    def append(self, value):
        list.append(self, value)
        self.check_type()
    
    def extend(self, iterable):
        list.extend(self, iterable)
        self.check_type()
    
    def insert(self, index, value):
        list.append(self, index, value)
        self.check_type()
    
    ######################
    # consistency checks #
    ######################
    
    def check_type(self):
        for elem in self:
            if not isinstance(elem, Motif):
                raise TypeError, "element %s is not of %s" % (elem, Motif)
    
    def check_unique(self):
        for i in xrange(len(self)):
            for j in xrange(i + 1, len(self)):
                if Motif.isduplicate(self[i], self[j]):
                    raise ValueError, "motifs %s and %s of %s are the same" % (i, j, self)
    
    #################
    # input methods #
    #################
    
    def from_xml_file(self, xml_file_name, schema_file_name=SCHEMA_FILE_NAME):
        """Read list of motifs from XML file and append to MotifList."""
        # parse motif XML
        tree = etree.parse(xml_file_name)
        # retrieve root tag
        root = tree.getroot()
        if root.tag != 'motifs':
            return  # error
        # extract motif data
        for child in root:
            new_motif = Motif()
            new_motif.from_xml_etree(child)
            self.append(new_motif)
    
    ##################
    # output methods #
    ##################
    
    def to_xml(self, pretty_print=False):
        """Write list of motifs to XML file."""
        # root <motifs> element
        root = etree.Element('motifs')
        # child <motif> elements
        for motif in self:
            motif_element = motif.to_xml_etree()
            root.append(motif_element)
        motifs_as_xml = etree.tostring(root)
        if pretty_print is True:
            motifs_as_xml = minidom.parseString(motifs_as_xml).toprettyxml(indent="  ")
        return motifs_as_xml
    

