#! /usr/bin/env python
from xml.etree.ElementTree import *
import xml.dom.minidom as minidom

def List2Str(l, con=";"):
    """use ';' to join list"""
    return con.join(l)

def PssmValidator(pssm):
    """validate PSSM format"""
    for pos in pssm:
        if len(pos)!=4:
            return 1
        for base in pos:
            try:
                float(base)
            except:
                return 1
    return 0

class MotifParser:
    """
    MotifParser (MP)
    function: parser motif xml file and return a MP Object "p", you can get the dictionary by "p.motifs"
    
    methods:
    __init__(self, xmlfile = '')   #Setup a MP Object, and parser the xmlfile.
    __add__(self, mp2)   #Add two MP into one, delete duplicates with same motif id. Return a MP object.
    __sub__(self, mp2)   #Subtract motifs of mp2 from the MP object. Return a MP object.
    Parser(self, xmlfile)   #Parser the xmlfile.
    ParserTable(self, tablefile)   #Parser from a "\t" splitted table txt file. Get a MP Object.
    ToXml(self, xmlfile, xsl=False)   #Output the MP Object to xmlfile. if xsl=True, also output the xsl file.
    ToTable(self, tabfile)   #Output the MP Object to table txt file.
    GetAttr(self, attr)    #Get all data of the attribute. Return a list.
    SearchMotif(self, **attrs)    #search a set of specific motifs in the MP Object. Return a MP Object.
    String(self, mid)   #Return a specific motif in a formatted string and readable type.
    
    xml file format:
    <motifs>
      <motif id='M00347' edition="2">                        #string
        <source>...</source>                                 #string
        <sourcefile>...</sourcefile>                         #string
        <status/>
        <specieslist><species>...</species></specieslist>    #string
        <fullname>...</fullname>                             #string
        <pmid>...</pmid>                                     #string
        <pssm><pos num='1'><A/><C/><G/><T/></pos>...</pssm>  #float
        <numseqs>...</numseqs>                               #string
        <symbol>...</symbol>                                 #string
        <synonymlist><synonym/>...</synonymlist>             #string
        <entrezlist><entrez/>...</entrezlist>                #string
        <refseqlist><refseq/>...</refseqlist>                #string
        <dbd>...</dbd>                                       #string
      </motif>
      <motif>
      </motif>
      ...
      </motifs>
      
    Details of the dictionary of MP object:
      'id'           list of string    identification in each db.
      'edition'      list of string    edition for JASPAR
      'source'       list of string    from which database
      'sourcefile'   list of string    url link to get the motif in original db
      'status'
      'numseqs'      list of string    length of motif 
      'pmid'         list of string    article support this motif
      'dbd'          list of string    DNA binding domain
      'description'  list of string    gene description
      'species'      list of strings   species the gene belong to
      'entrez'       list of strings   entrezs of the gene
      'symbol'       list of string    symbol of the gene
      'synonym'      list of strings   non-official gene symbols of the gene
      'refseq'       list of strings   refseqs of the gene
      'pssm'         2d-list of float  pssm matrix of the motif
    """
    def __init__(self, xmlfile = ''):
        """MP = MotifParser()
        Setup a MP Object, and parser the xmlfile."""
        self.motifs = {}
        self.temp_print = ''
        self.attr_list = ['id','edition','source', 'sourcefile', 'status', 'numseqs', 'pmid', 'dbd', \
        'description', 'species', 'entrez', 'symbol', 'synonym', 'refseq', 'pssm']
        self.attr_info = {'id':'Motif ID', 'edition':'Edition', 'source':'Source', 'sourcefile':'Source Link', 'status':'Status', \
        'numseqs':'Motif Length', 'pmid':'Pubmed ID', 'dbd':'DNA Binding Domain', 'description':'Gene Description', \
        'species':'Species', 'entrez':'Entrez GeneID', 'symbol':'Gene Symbol', \
        'synonym':'Non Official Synonyms', 'refseq':'Refseq Accession', 'pssm':'PSSM matrix'}
        
        if xmlfile:
            self.Parser(xmlfile)
    
    def Parser(self, xmlfile):
        """MP.Parser(xmlfile)
        Parser the xmlfile."""
        xmltree = ElementTree()
        try:
            xmltree.parse(xmlfile)
        except IOError:
            print "Fail to parser the xml file. Not exist or format error?"

        for pos in xmltree.findall("motif"):
            tag = 'id'
            attrib = pos.attrib
            id = attrib[tag]
            #print id
            self.motifs[id] = {}
            self.motifs[id][tag] = [id]
            
            tag = 'edition'
            try:
                self.motifs[id][tag] = [attrib[tag]]
            except:
                self.motifs[id][tag] = []
    
            for tag in ('source', 'sourcefile', 'status', 'description', 'numseqs', 'pmid', 'dbd', 'symbol'):
                if pos.find(tag)!=None:
                    self.motifs[id][tag] = [pos.find(tag).text.strip()]
                else:
                    self.motifs[id][tag] = []
                    
            for tag in ('species', 'entrez', 'synonym', 'refseq'):
                if pos.find(tag+'list')!=None:
                    t = pos.find(tag+'list')
                    self.motifs[id][tag] = [i.text.strip() for i in t.findall(tag)]
                else:
                    self.motifs[id][tag] = []
                    
            for tag in ('pssm',):
                self.motifs[id][tag]=[]
                if pos.find(tag)!=None:
                    plist = pos.find(tag).findall('pos')
                    plist.sort()
                    for t in plist:
                        t = t.findall('*')
                        t.sort()
                        self.motifs[id][tag].append([ float(s.text.strip()) for s in t ])
                        
    def ParserTable(self, tfile):
        """Parser from a "\t" splitted table txt file. 
        The first col should be col name and in lowercase letter.
        The pssm format like this: [[0.2,0.3,0.3,0.2],[0.1,0.8,0.05,0.05]]"""
        inf = open(tfile)
        line = inf.readline()
        headList = line.rstrip('\n').split('\t')
        headIndex = {}
        for i in range(len(headList)):
            headIndex[headList[i]] = i
        for each in headList:
            if each not in self.attr_list:
                print "Error: column name '%s' is not a node, use 'MP_Object.attrlist' to get formal format." %each
                return 1
                
        for line in inf:
            linel = line.rstrip('\n').split('\t')
            tag = 'id'
            id = linel[headIndex[tag]]
            self.motifs[id] = {}
            self.motifs[id][tag] = [id]
            
            tag = 'edition'
            self.motifs[id][tag] = []
            if linel[headIndex[tag]]:
                self.motifs[id][tag] = [linel[headIndex[tag]]] 
                
            for tag in ('source', 'sourcefile', 'status', 'description', 'numseqs', 'pmid', 'dbd', 'symbol'):
                self.motifs[id][tag] = []
                if linel[headIndex[tag]]:
                    self.motifs[id][tag] = [linel[headIndex[tag]]]

            for tag in ('species', 'entrez', 'synonym', 'refseq'):
                self.motifs[id][tag] = []
                if linel[headIndex[tag]]:
                    self.motifs[id][tag] = linel[headIndex[tag]].split(";")
                    
            for tag in ('pssm',):
                pssm = []
                if linel[headIndex[tag]]:
                    exec('pssm=%s' %linel[headIndex[tag]])
                    if PssmValidator(pssm)!=0:
                        print "pssm format error. id: %s" %id
                self.motifs[id][tag] = pssm
        print "Success parser from table."
                        
    def GetAttr(self, attr):
        """MP.GetAttr(attr)
        Get all data of the attribute / tag. Return a list."""
        if attr not in self.attr_list:
            print "Wrong input attr, select attr from: \n:",List2Str(self.attr_list)
            return ''
        if attr == 'pssm':
            print "Not support to get pssm. you can use MP.ToTable() ;)\n"
            return 1
        res = []
        for i in self.motifs.values():
            res.extend(i[attr])
        res.sort()
        return List2Str(res)
        
    def SearchMotif(self, **attrs):
        """search a set of specific motifs in the MP Object. Return a MP Object
        choose arguments from
        e.g) MP2 = MP.SearchMotif(species="Homo sapiens",source="JASPAR")"""
        #print attrs
        for i in attrs.keys():
            if i not in self.attr_list:
                print "Wrong input attr:%s, select attr from:\n:" %i, List2Str(self.attr_list)
                return ''
        sub_motifs = MotifParser()
        sub_motifs.motifs = self.motifs
        for attr in attrs.items():
            temp_dict = {}
            for i in sub_motifs.motifs.items():
                if attr[1] in i[1][attr[0]]:
                    temp_dict[i[0]] = i[1]
            sub_motifs.motifs = temp_dict
        print "Extract %d records."%len(sub_motifs.motifs.keys())
        return sub_motifs

    def String(self, mid):
        """Return a specific motif in a formatted string and readable type.
        e.g) MP.String("M00913")"""
        try:
            t_motif = self.motifs[mid]
        except:
            print "ID incorrect, can't find this Motif ID."
        motif_string = ['\n']
        for tag in self.attr_list[:-1]: #remove tag 'pssm'
            if t_motif[tag]:
                motif_string.append("%s: %s\n" %(self.attr_info[tag], List2Str(t_motif[tag]) ))
            else:
                motif_string.append("%s: None\n" %(self.attr_info[tag], ))

        tag = 'pssm'
        if t_motif[tag]:
            motif_string.append("PSSM:        A      C      G      T\n")
            for i in range(len(t_motif[tag])):
                motif_string.append("|%6d"%(i+1,) + "  %3.3f  %3.3f  %3.3f  %3.3f\n" %tuple(t_motif[tag][i]))
        else:
            motif_string.append("%s: None\n" %self.attr_info[tag])
            
        return List2Str(motif_string,"")

    def __add__(self, mp2):
        """MP1.__add__(MP2) <==> MP1+MP2
        Add two MP into one, delete duplicates with same motif id. Return a MP object."""
        res_motifs = MotifParser()
        res_motifs.motifs = self.motifs
        motif_id_list = res_motifs.motifs.keys()
        for i in mp2.motifs:
            if i not in motif_id_list:
                res_motifs.motifs[i] = mp2.motifs[i]
        return res_motifs
    
    def __sub__(self, mp2):
        """MP1.__sub__(MP2) <==> MP1-MP2
        Subtract motifs of mp2 from the MP object. Return a MP object."""
        res_motifs = MotifParser()
        res_motifs.motifs = self.motifs
        motif_id_list = res_motifs.motifs.keys()
        for i in mp2.motifs:
            if i in motif_id_list:
                del(res_motifs.motifs[i])
        return res_motifs

    def ToXml(self, xmlfile, xsl=False):
        """MP.ToXML(xmlfile)
        Output the MP Object to xmlfile."""
        doc = minidom.Document()
        motifs = doc.createElement("motifs")
        doc.appendChild(motifs)
        
        t_motifs = self.motifs.values()
        t_motifs.sort(key=lambda x:x['id'])
        for mo in t_motifs:
            # Create the main element
            motif = doc.createElement("motif")
            # Create main Attrs
            motif.setAttribute('id', List2Str(mo['id']))   #set id
            if  mo['edition']:   #set edition
                motif.setAttribute('edition', List2Str(mo['edition']))
            motifs.appendChild(motif)
    
            #Create elements
            for tag in ('source', 'sourcefile', 'status', 'description', 'numseqs', 'pmid', 'dbd', 'symbol'):
                if mo[tag]: #skip to setup the node if  it is None or '-'
                    element = doc.createElement(tag)
                    motif.appendChild(element)
                    ptext = doc.createTextNode(List2Str(mo[tag]))
                    element.appendChild(ptext)   
            for tag in ('species', 'entrez', 'synonym', 'refseq'):
                if mo[tag]:
                    element = doc.createElement(tag+"list")
                    motif.appendChild(element)
                    for items in mo[tag]:
                        subelement = doc.createElement(tag)
                        element.appendChild(subelement)
                        ptext = doc.createTextNode(items)
                        subelement.appendChild(ptext)
            for tag in ("pssm",):
                if mo[tag]:
                    baseIndex = ['A', 'C', 'G', 'T']
                    pssm = doc.createElement(tag)
                    motif.appendChild(pssm)
                    for i in range(len(mo[tag])):
                        pos = doc.createElement("pos")
                        pos.setAttribute("num", "%d" %(i+1,))
                        pssm.appendChild(pos)
                        for j in range(4):
                            t = doc.createElement(baseIndex[j])
                            pos.appendChild(t)
                            ptext = doc.createTextNode(str(mo[tag][i][j]))
                            t.appendChild(ptext)
        xmlString = doc.toprettyxml(indent="  ")
        xmlString = xmlString.split("\n",1)
        
        #output to xmlfile
        xmlfile_ = xmlfile.rstrip(".xml")
        outf = open(xmlfile_+".xml",'w')
        outf.write(xmlString[0]+"\n")
        outf.write(r'<?xml-stylesheet type="text/xsl" href="%s.xsl"?>'% xmlfile_ +"\n")
        outf.write(xmlString[1])
        outf.close()
        print "Output xml to file: %s.xml." %xmlfile_
        
        if xsl:
            outf = open(xmlfile_+".xsl",'w')
            outf.write("""\
<?xml version="1.0"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method="html" encoding="utf-8"/>
<xsl:template match="/">

<html>
<head>
	<title>%s</title>
</head>
<body>
<table cellspacing="1" border="1" bordercolor="#FF9900">
	<THEAD>
		<TR>
			<TD bgcolor="#FFFF99"><B>ID</B></TD>
			<TD bgcolor="#FFFF99"><B>SOURCE</B></TD>
			<TD bgcolor="#FFFF99"><B>STATUS</B></TD>
			<TD bgcolor="#FFFF99"><B>SPECIES</B></TD>
			<TD bgcolor="#FFFF99"><B>SYMBOL</B></TD>
			<TD bgcolor="#FFFF99"><B>ENTREZ</B></TD>
			<TD bgcolor="#FFFF99"><B>REFSEQ</B></TD>
			<TD bgcolor="#FFFF99"><B>DESCRIPTION</B></TD>
			<TD bgcolor="#FFFF99"><B>DBD</B></TD>
			<TD bgcolor="#FFFF99"><B>PMID</B></TD>
			<TD bgcolor="#FFFF99"><B>SEQLOGO</B></TD>
		</TR>
	</THEAD>
	<TBODY>
		<xsl:for-each select="motifs/motif">
		<TR>
			<TD><a target="_blank"><xsl:attribute name="href"><xsl:value-of select="normalize-space(sourcefile)" /></xsl:attribute><xsl:value-of select="@id" /></a></TD>
			<TD><xsl:if test="not(source)">-</xsl:if>	<xsl:value-of select="source" /></TD>
			<TD><xsl:if test="not(status)">-</xsl:if>	<xsl:value-of select="status" /></TD>
			<TD><xsl:if test="not(specieslist)">-</xsl:if>	<xsl:value-of select="specieslist" /></TD>
			<TD><xsl:if test="not(symbol)">-</xsl:if>	<xsl:value-of select="symbol" /></TD>
			<TD><xsl:if test="not(entrezlist)">-</xsl:if>	
			  <xsl:for-each select="entrezlist/entrez">
			    <a target="_blank"><xsl:attribute name="href">http://www.ncbi.nlm.nih.gov/gene/?term=<xsl:value-of select="normalize-space(.)" /></xsl:attribute><xsl:value-of select="." /></a>
			  </xsl:for-each></TD>
			<TD><xsl:if test="not(refseqlist)">-</xsl:if>	<xsl:value-of select="refseqlist" /></TD>
			<TD><xsl:if test="not(description)">-</xsl:if>	<xsl:value-of select="description" /></TD>
			<TD><xsl:if test="not(dbd)">-</xsl:if>		<xsl:value-of select="dbd" /></TD>
			<TD><xsl:if test="not(pmid)">-</xsl:if>		<a target="_blank"><xsl:attribute name="href">http://www.ncbi.nlm.nih.gov/pubmed/?term=<xsl:value-of select="normalize-space(pmid)" /></xsl:attribute><xsl:value-of select="pmid" /></a></TD>
			<TD><xsl:if test="not(pssm)">-</xsl:if>		<a href="seqLogo/{@id}.png" target="_blank"><img width="180" height="90" src="seqLogo/{@id}.png"></img></a></TD>
		</TR>
		</xsl:for-each>
	</TBODY>
</table>
</body>
</html>
</xsl:template>
</xsl:stylesheet>"""%xmlfile)
            outf.close()
            
    def ToTable(self, tabfile):
        outf = open(tabfile,'w')
        outf.write(List2Str(self.attr_list, "\t")+"\n")
        t_motifs = self.motifs.values()
        t_motifs.sort(key=lambda x:x['id'])
        for each in t_motifs:
            motif = [List2Str(each[t]) for t in self.attr_list[:-1]]
            motif.append(str(each['pssm']))
            outf.write(List2Str(motif, "\t")+"\n")
        outf.close()
        
    def ConvertToYing(self):
        import base
        import numpy
        motiflist = base.MotifList()
        for each in self.motifs.values():
            pm = base.Motif()
            pm.id = each['id'][0]
            pm.status = None
            if each['source']:
                pm.source = each['source'][0]
            if each['sourcefile']:
                pm.sourcefile = each['sourcefile'][0]
            if each['species']:
                pm.species = each['species']
            if each['entrez']:
                pm.entrez = [int(t) for t in each['entrez']]
            if each['symbol']:
                pm.symbol = each['symbol']
            if each['synonym']:
                pm.synonyms = each['synonym']
            if each['description']:
                pm.fullname = each['description'][0]
            if each['dbd']:
                pm.dbd = each['dbd'][0]
            if each['pmid']:
                pm.pmid = each['pmid'][0]
            if each['pssm']:
                pm.pssm = numpy.array(each['pssm'], float)
                if pm.pssm.shape[1] != 4:
                    raise ValueError, "motif PSSM must have 4 columns"
            pm.numseqs = None
            pm.curators = []
            pm.results = None
            pm.antisense = False # reverse complements need to be taken to make tree consistent
            motiflist.append(pm)
        return motiflist
        

