"""
Copyright (c) 2010 Cliff Meyer <cliff@jimmy.harvard.edu>
This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included with
the distribution).
contact: cliff@jimmy.harvard.edu
"""

import sys, os, time, math
import operator, subprocess, optparse
import numpy
import numpy.linalg as LA
import seq.count as cnt
import cistrome.regions as cis
import motifIO
import motifIO.base
import seq._seq
import random
import basicstat as bs

LEN   = 1000
NLEN  = 200
MEANPROB = seq._seq.MEAN_OPTION
MAXPROB  = seq._seq.MAX_OPTION
RVS      = seq._seq.RVS
FWD      = seq._seq.FWD
MAXSEQS  = seq._seq.MAXMOTIFS
BIGVAL   = 9e9

def conservation_mask(A,cons_cutoff=0.90):
    A.getPhast()
    masklocs = []
    for i,elem in enumerate(A.conservation):
        masklocs += [ (i,j,j+1) for j,score in enumerate(elem) if score < cons_cutoff ]
    #print masklocs
    A.getSequence()
    cis.maskseq( A.seq, masklocs )

def header_check(header,id):
    if type(id) != type([]):
        raise TypeError("id is supposed to be a list.") 
    for i,elem in enumerate(id):
        try:
            hidx = header[elem]
        except KeyError:
            print >> sys.stderr,'Key not found - searching for partial match', elem
            try:
                for keyelem in header.keys():
                    if keyelem.find(elem) >= 0:
                        hidx = header[keyelem]
                        header[elem] = hidx
            except:
                print >> sys.stderr,'Partial match to key not found', elem
                raise
    return

def parse_table( tablefile,genome='hg18',annot_id=None,select_id=None,select_L=[-BIGVAL],select_U=[BIGVAL],nuc_id=None,lin_id=None,prommark_id=None,enhmark_id=None ):

    """
    annot_id  is a list (or None) for reading annotation columns into info (a list of lists one for each region)
    select_id is a list (or None) for reading selection  columns
    select_L and select_U are lists defining lower and upper bounds for regions to be included in analysis
    nuc_id  is a list of columns with tag counts in nucleosome regions
    lin_id  is a list of columns with tag counts in linker regions. lin_id entries are paired with nuc_id entries.
    """

    try:
        fp = open(tablefile,'r')
    except IOError:
        print >> sys.stderr,"Error trying to open %s." % tablefile
        raise

    chrom,start,end  = [],[],[]
    nuc = []
    lin = []

    header = {}
    while 'start' not in header:
        line  = fp.readline()
        field = line.split()
        if field[1] == 'start':
            for i,elem in enumerate(field):
                header[elem] = i

    P = cis.interval(genome,"NSD table")
    P.genome = genome
    P.chrom, P.start, P.end, P.length  = [],[],[],[]
    P.strand, P.mid, P.name = [],[],[]

    # check annot_id, select_id, nuc_id and lin_id are in header
    if annot_id != None and len(annot_id) > 0:
        header_check(header,annot_id)
        P.info = []
    else:
        P.info = None

    if nuc_id != None:
        header_check(header,nuc_id)
        P.nuc = []
    else:
        P.nuc = None

    if lin_id != None:
        header_check(header,lin_id)
        P.lin = []
    else:
        P.lin = None

    if select_id != None:
        header_check(header,select_id)
        for k,elem in enumerate(select_id):
            print "filtering on field:%s (%f,%f)" % (select_id[k],select_L[k],select_U[k])

    if prommark_id != None:
        header_check(header,prommark_id)
        P.prom = []
    else:
        P.prom = None

    if enhmark_id != None:
        header_check(header,enhmark_id)
        P.enh = []
    else:
        P.enh = None

    for count,line in enumerate( fp.readlines() ):

        if ( line[0:3] == 'chr' ):
            t = line.split()
 
            select_flag = True
            if select_id != None:

                if type(select_id) != type([]):
                    raise TypeError("Select_id is supposed to be a list.") 
                if type(select_U) != type([]):
                    raise TypeError("Select_U is supposed to be a list.") 
                if type(select_L) != type([]):
                    raise TypeError("Select_L is supposed to be a list.") 

                select_flag = True
                for i,elem in enumerate(select_id):
                    #print elem, header[elem], t[header[elem]], select_L[i]
                    #print elem, header[elem], t[header[elem]], select_U[i]
                    try:
                        hidx = header[elem]
                    except KeyError:
                        print >> sys.stderr,'Key not found', select_id
                        raise

                    if ( float(t[hidx]) < select_L[i] ) or ( float(t[hidx]) >= select_U[i] ):
                        select_flag = False
 
            #if select_id == None or ( ( float( t[header[select_id]] ) >= select_L ) and ( float( t[header[select_id]] ) < select_U ) ): 
            if select_flag == True:
                # match seqs with chrom locations !
                P.chrom.append( t[0] )
                P.start.append( int(t[1]) )
                P.end.append( int(t[2]) )
 
                P.nuc.append( [ float( t[ header[elem] ] ) for elem in nuc_id ] )
                if lin_id:
                    P.lin.append( [ float( t[ header[elem] ] ) for elem in lin_id ] )
                if prommark_id:
                    P.prom.append( [ float( t[ header[elem] ] ) for elem in prommark_id ] )
                if enhmark_id:
                    P.enh.append( [ float( t[ header[elem] ] ) for elem in enhmark_id ] )
                if annot_id:
                    P.info.append( [ float( t[ header[elem] ] ) for elem in annot_id ] )

    P.mid,P.length = [],[]
    for i,elem in enumerate( P.chrom ):
        P.length.append( P.end[i] - P.start[i] )
        P.mid.append( int( 0.5*( P.start[i] + P.end[i] ) ) )

    P.strand = len(P.chrom)*['+']
    P.name = [ str(x) for i,x in enumerate(P.chrom) ]
    P.description = tablefile

    fp.close()

    print >> sys.stdout, "Summary of parsed file"
    print >> sys.stdout, "number of regions %d" % len(P.chrom)
    print >> sys.stdout, "number of samples for nucleosomes %d" % len(P.nuc[0])

    return P


def calc_NSD(P,idx):
    """
    calculates NSD score for each regions.
    idx is index of treatment sample
    method of calculation:
        if P.lin data is supplied linker counts are subtracted
        if P.nuc has > 1 the NSD score is computed relative to the mean of the controls 
    """

    nuc, lin, seqs = P.nuc, P.lin, P.seq

    if P.lin != None and len(P.lin) > 0:
        lin_flag = True
    else:
        lin_flag = False

    nuc = numpy.array(nuc)
    if lin_flag:
        lin = numpy.array(lin)

    nd = nuc.shape[1]

    # total number of tags in flanking and linker regions together 
    if lin_flag:
        sumt = [ nuc[:,i].sum() + lin[:,i].sum() for i in range(nd) ]
    else:
        sumt = [ nuc[:,i].sum() for i in range(nd) ]

    # fraction of tag counts relative to total counts in all regions
    for i in range(nd):
        nuc[:,i] = nuc[:,i]/sumt[i]
        if lin_flag:
            lin[:,i] = lin[:,i]/sumt[i]

    # square root of normalized tag counts 
    nuc = numpy.sqrt(nuc)
    if lin_flag:
        lin = numpy.sqrt(lin)

    # subtract treatment sample linker value if requested
    if lin_flag:
        NSD = ( nuc[:,idx] - lin[:,idx] ) 
    else:
        NSD = ( nuc[:,idx] ) 

    # subtract mean of controls
    for j in range(nd):
        if j != idx:
            if lin_flag:
                NSD -= (1.0/float( nd - 1 ))*( nuc[:,j] - lin[:,j] ) 
            else:
                NSD -= (1.0/float( nd - 1 ))*( nuc[:,j] ) 


    NSDmu  = NSD.mean()
    NSDsd  = NSD.std()
    print "mean and sd", NSDmu, NSDsd
    NSD = (NSD-NSDmu)/NSDsd
    return NSD


def binomial_stat( s_idx_X, score_X, s_idx_Y, score_Y, n_X, n_Y, cutoff = 5.0 ):
    """
    compute the binomial pvalue for motifs above in X being greater than those in Y
    max 1 motif counted in each sequence
    if there are less than 10 hits pvalue is set to -1
    """

    MINHITS         = 10
    count_X,count_Y = 0,0 
    dict_X,dict_Y   = {},{}

    for i,elem in enumerate(s_idx_X):
        if elem not in dict_X and score_X[i] > cutoff:
            count_X += 1
            dict_X[elem] = 1

    for i,elem in enumerate(s_idx_Y):
        if elem not in dict_Y and score_Y[i] > cutoff:
            count_Y += 1
            dict_Y[elem] = 1

    p = float(count_Y)/n_Y
    pval = bs.binomial_cdf( max(0,count_X-1), n_X, p, lower=False ) 

    # if there are too few hits set pvalue to -1
    if count_X < MINHITS:
        pval = -1

    return {'numhits_fg':count_X, 'numhits_bg':count_Y, 'cutoff':cutoff, 'pvalue':pval, 'n':n_X }


def seqpos_stat( length, start, end, score, offset ):

    start   = numpy.array(start,float)
    end     = numpy.array(end,float)
    length  = numpy.array(length,float)

    MINHITS = 20
    score   = -score
    frac    = numpy.abs( ( start + end - offset )/(length-offset) - 1.0 )

    idx     = score.argsort()
    mu_bias = 0.5 

    cumfrac, meanfracpos_t, zscore_t = 0,[],[]
    zscore_min, kmin, numhits, cuthits, meanpos = 99, -1, 0, 0, 0

    for numhits,i in enumerate(idx):

        if frac[i] > 1.0 or frac[i] < 0.0:
            #print "frac out of bounds", start[i], end[i], length[i], frac[i]
            raise(ValueError,"Fraction out of bounds %4.2f" % frac[i] )

        cumfrac += frac[i]
        meanfracpos_t.append( cumfrac/numhits )
        if numhits > MINHITS and -score[i] > 0.0:
            zscore = ( cumfrac - numhits*mu_bias )/( math.sqrt( numhits/12.0 ) )
            #print 'zzz', numhits,score[i],frac[i],meanfracpos_t[numhits],zscore
            if zscore < zscore_min:
                zscore_min = zscore
                kmin = -score[i]
                cuthits = numhits
                meanpos = cumfrac/numhits - 0.5

    if cuthits > MINHITS:
        # NOTE: w/ the pvalue, we are assuming the normal distribution
        # get the raw p-value score
        z_mean_adj = - ( motifIO.base.MUALPHA0 + motifIO.base.MUALPHA1*math.log( math.log( numhits ) ) )
        z_sd_adj = motifIO.base.SDALPHA0 + motifIO.base.SDALPHA1*math.log( math.log( numhits ) )
        zscore_min_adjusted = ( zscore_min - z_mean_adj )/z_sd_adj  
        pvalue = max(1e-30, bs.normal_cdf(zscore_min_adjusted))
        #print "#----->", cuthits, z_mean_adj, z_sd_adj, zscore_min_adjusted, pvalue
    else:
        zscore_min_adjusted = 0
        pvalue = 1.0
        #M = 0.25*numpy.ones( self.pssm.shape )

    results = {
        'numhits': cuthits,
        'meanposition': meanpos,
        'cutoff': kmin,
        'zscore': zscore_min_adjusted,
        'pvalue': pvalue,
        #'pssm': M
    }

    return results


def NSD_select(P,idx,suffix,nfrac=0.01,bias=False,printnsd=False):
    """
    Select the regions to be analysed in seqpos
    ntop is the fraction of sites to be considered (0,0.5]
    If .xls and ntop is given also return a background set
    """
 
    # for .xls files lin and nuc fields are expected and the NSD score is computed
    if suffix == 'xls':
 
        NSD   = calc_NSD(P,idx)

        if bias == True:
            NSD = GC_bias_correction(P,NSD)
        P.val = NSD

        ntot  = len(P.val)
        ntop  = int(nfrac*ntot)
        P.rank( Field='val', order='decreasing' )
        lin,nuc = P.lin,P.nuc
        P.lin,P.nuc = None,None
 
        if ntop:
            print "selecting top %d out of %d regions" %  (ntop,ntot)
            X = P.Select( Field='idx', UpperBound=ntop, LowerBound=0 )
            ntot = len(P.chrom)
            Y = P.Select( Field='idx', UpperBound=( ntot + ntop )/2, LowerBound=(ntot - ntop)/2 )
        else:
            X = P
            Y = None

        P.lin,P.nuc = lin,nuc

    # for.bed file values are from the file itself
    else:
        
        if nfrac:
            if nfrac < 1.0:
                ntot  = len(P.val)
                ntop  = int(nfrac*ntot)
                P.rank( Field='val', order='decreasing' )
                X = P.Select( Field='idx', UpperBound=ntop, LowerBound=0 )
            else:
                X = P
        else:
            X = P

        Y = None

    # print NSD scores of selected regions
    if printnsd == True:
        for k_idx,elem in enumerate(X.val):
            print "%s\t%d\t%d\t%6.3f" % ( X.chrom[k_idx], X.start[k_idx], X.end[k_idx], 1e6*elem )
  
    return X,Y


def bed_or_table_parse( tablefile, genome, annot_id, select_id, select_L, select_U, nuc_id, lin_id, prommark_id, enhmark_id ):

    filename = os.path.split(tablefile)[1]
    suffix   = filename.split('.')

    try:
        suffix = suffix[-1]
    except IndexError:
        print >>sys.stderr,"%s must be a .bed file or a .xls file" % tablefile
        raise usage 

    if suffix == 'xls': 
        P = parse_table( tablefile, genome=genome, annot_id=annot_id, select_id=select_id, select_L=select_L, select_U=select_U, nuc_id=nuc_id, lin_id=lin_id, prommark_id=prommark_id, enhmark_id=enhmark_id )
    elif suffix == 'bed':
        try:
            nuc_id.append("bed")
        except:
            nuc_id = ["bed"]
        P = cis.interval(genome) 
        P.read_bed(tablefile)
    else:
        print >>sys.stderr,"%s must be a .bed file or a .xls file" % tablefile
        raise usage

    return suffix,P


def main_seqpos( tablefile, known_motifs, names=None, genome='hg18', span=LEN, select_id = None, select_U = 9e9, select_L = -9e9, nuc_id=None, lin_id=None, prommark_id=None, score_option='max', ntop=None, output_file=None, bias=False, printnsd=False, analysis = 'pos', order=1, debug=False ): 

    """
    nuc_id is list of column names for nucleosome tag counts
    lin_id is list of column names for linker tag counts
    score_option can be 'max' or 'mean'. This is the motif sequence score associated with a region.
    if suffix of tablefile is xls the file is processed as a headed file of tag counts
    if suffix of tablefile is bed the file is processed as a bed file
    analysis: 'pos' or 'enrich' 
    """

    MOTIFMIN = 1e-3
    #order    = 2 #0,1,2

    if names and len( names ) > 0:
        test_motifs = [ elem for elem in known_motifs if elem.id in names ]
    else:
        test_motifs = known_motifs

    suffix,P = bed_or_table_parse( tablefile, genome, None, select_id, select_L, select_U, nuc_id, lin_id, None, None )
    #print suffix
    #print nuc_id
    
    width = P.length[:]
    P.TrimBed(length=span)
    P.getSequence()

    # trim sequence if file is xls (assume this is table of seq tag counts)
    # cut nucleosome regions out of sequence.
    if analysis == 'pos':

        if suffix == 'xls':
            for i,elem in enumerate(width):
                pad = int( ( width[i] - P.length[i] )/2.0 )
                if pad < 0:
                    raise( ValueError, "Length of sequence %d %d" % (width[i],P.length[i]) )

                P.seq[i] = P.seq[i][0:pad] + P.seq[i][pad+NLEN:width[i]-pad-NLEN] + P.seq[i][width[i]-pad:width[i]] 
        tsorted_X = {}
        tsorted_Y = {}
    elif analysis == 'enrich':
        tsorted_X = {}

    P.checkseq()
    bgseq = cnt.count(P.seq)

    if output_file:
        try:
            ofile = open(output_file,"w")
        except IOError:
            print >>sys.stderr, "Unable to open ", output_file
            return None
    else:
        ofile = None

    for idx,treat_name in enumerate(nuc_id):
        
        if debug:
            print treat_name

        X,Y     = NSD_select(P,idx,suffix,ntop,bias=bias,printnsd=printnsd)
        bgseq_X = cnt.count(X.seq)
 
        if Y:
            bgseq_Y = cnt.count(Y.seq)

        if analysis == 'pos':
            tstat_X = []
            tstat_Y = []
        elif analysis == 'enrich':
            bn_stat = []

        for motif_elem in test_motifs:
 
            pssm = motif_elem.pssm
            id   = str(motif_elem.id)

            if motif_elem.symbol:
                sym = motif_elem.symbol[0].strip()
                sym = sym.replace( '\/','_' )
            else:
                sym = id
                sym = sym.replace( '\/','_' )

            if analysis == "pos":

                if X:
                    s_idx,start,end,orient,score = seq._seq.seqscan( X.seq, pssm, bgseq_X, order, MAXPROB )

                    if debug:
                        print "hits for: ", id
                        for i,elem in enumerate(s_idx):
                            if orient[i] == RVS:
                                print X.chrom[elem], X.start[elem] + start[i], X.start[elem] + end[i], orient[i], score[i], cnt.reverse_complement( X.seq[elem][ max(0,start[i]-order): min( len(X.seq[elem]), end[i]+order) ] )
                            else:
                                print X.chrom[elem], X.start[elem] + start[i], X.start[elem] + end[i], orient[i], score[i], X.seq[elem][ max(0,start[i]-order): min(len(X.seq[elem]),end[i]+order) ] 

                    lengths = []
                    for i,elem in enumerate(s_idx):
                        lengths.append( len(X.seq[elem]) )
                    #----------------- adjust score ---------------------
                    score = ( numpy.log( score + MOTIFMIN ) )
                    #----------------------------------------------------
                    results = seqpos_stat(   numpy.array(lengths), start, end, score, len(pssm) ) 
                    tstat_X.append( [ id, sym, motifIO.base.pwm_to_iupac(pssm), results['numhits'], results['meanposition'], results['cutoff'], results['zscore'], results['pvalue'] ] )

                if Y:
                    s_idx,start,end,orient,score = seq._seq.seqscan( Y.seq, pssm, bgseq_Y, order, MAXPROB )
                    lengths = []
                    for i,elem in enumerate(s_idx):
                        lengths.append( len(Y.seq[s_idx[elem]]) )
                    #----------------- adjust score ---------------------
                    score = ( numpy.log( score + MOTIFMIN ) )
                    #----------------------------------------------------
                    results = seqpos_stat( numpy.array(lengths), start, end, score, len(pssm) ) 
                    tstat_Y.append( [ id, sym, motifIO.base.pwm_to_iupac(pssm),  results['numhits'], results['meanposition'], results['cutoff'], results['zscore'], results['pvalue'] ] )
 
            elif analysis == "enrich": 

                s_idx_X,start_X,end_X,orient_X,score_X = seq._seq.seqscan( X.seq, pssm, bgseq_X, order, MAXPROB )
                s_idx_Y,start_Y,end_Y,orient_Y,score_Y = seq._seq.seqscan( Y.seq, pssm, bgseq_Y, order, MAXPROB )
                #----------------- adjust score ---------------------
                score_X = ( numpy.log( score_X + MOTIFMIN ) )
                score_Y = ( numpy.log( score_Y + MOTIFMIN ) )
                #----------------------------------------------------
                results = binomial_stat( s_idx_X, score_X, s_idx_Y, score_Y, len(X.seq), len(Y.seq) ) 
                bn_stat.append( [ id, sym, motifIO.base.pwm_to_iupac(pssm), results['numhits_fg'], results['numhits_bg'], results['n'], results['cutoff'], results['pvalue'] ] )

        if analysis == "pos":
            tsorted_X[treat_name] = sorted(tstat_X,key=operator.itemgetter(6),reverse=False)
            if Y:
                tsorted_Y[treat_name] = sorted(tstat_Y,key=operator.itemgetter(6),reverse=False)
        elif analysis == "enrich":
            tsorted_X[treat_name] = sorted(bn_stat,key=operator.itemgetter(7),reverse=False)

        if ofile:

            ofile.write("Analysis: %s\n" % analysis )
            ofile.write("Input file name: %s\n" % tablefile)
            ofile.write("Genome version: %s\n" % genome)
            ofile.write("Sequence width: %d\n" % span)
            ofile.write("Background Markov model: %d\n" % order )

            if select_id:
                ofile.write( "Filter fields: %s\n" % ','.join(select_id) ) 
                ofile.write( "Filter lower bounds: %s\n" % ','.join([str(x) for x in select_L]) )
                ofile.write( "Filter upper bounds: %s\n" % ','.join([str(x) for x in select_U]) )

            if nuc_id and len(nuc_id) > 0:
                ofile.write( "Nucleosome fields: %s\n" % ','.join([elem for elem in nuc_id if elem != None])) 
            
            if lin_id and len(lin_id) > 0:
                ofile.write( "Linker fields: %s\n" % ','.join([elem for elem in lin_id if elem != None]))

            #ofile.write( "Sequence score option: %s\n" % score_option )
            #ofile.write( "Number of regions in analysis: %d\n" % len(X.chrom) )
            ofile.write( "Treatment sample: %s\n" % treat_name )
 
            if analysis == "pos":

                ofile.write( "Number of treatment regions: %d\n" % len(X.seq) )
                ofile.write("\n-------------------------------------------------------------------------\n")
                ofile.write("ID\tsym\tconsensus\tnumhits\tmean\tcutoff\tzscore\tpval\n")

                if tsorted_X[treat_name]:
                    for elem in tsorted_X[treat_name]:

                        try:
                            print >> ofile, "%s\t%s\t%s\t%5d\t%4.2f\t%4.2f\t%4.2f\t%4.2e" % (elem[0],elem[1],elem[2],elem[3],elem[4],elem[5],elem[6],elem[7])
                        except:
                            print "ERR",elem
                            pass

                    ofile.write(  "-------------------------------------------------------------------------\n\n")
        
            if analysis == "enrich":

                ofile.write( "Number of treatment regions: %d\n" % len(X.seq) )
                ofile.write( "Number of control regions:   %d\n" % len(Y.seq) )

                ofile.write("\n-------------------------------------------------------------------------\n")
                ofile.write("ID\tsym\tconsensus\tnumhits_fg\tnumhits_bg\tn\tcutoff\tpval\n")

                if tsorted_X[treat_name]:
                    for elem in tsorted_X[treat_name]:

                        try:
                            print >> ofile, "%s\t%s\t%s\t%5d\t%5d\t%5d\t%5d\t%4.2e" % (elem[0],elem[1],elem[2],elem[3],elem[4],elem[5],elem[6],elem[7])
                        except:
                            print "ERR",elem
                            pass

                    ofile.write(  "-------------------------------------------------------------------------\n\n")

    if ofile:
        ofile.close()

    return tsorted_X


def print_NSD( tablefile, known_motifs, genome='hg18', select_id = None, select_U = 9e9, select_L = -9e9, nuc_id=None, lin_id=None, score_option='max', posterior_cutoff=0.99, conservation_cutoff=0.90, output_file=None, bias=False, names=None ): 
    NSD,PS = calc_NSD(P,idx)
    if bias == True:
        for i,elem in enumerate(basecont):
            print '\t'.join( str(x) for x in elem ), NSD[i]
            X = numpy.array( [ [1,elem[1] + elem[2]] for elem in basecont ] ) 
            NSD = fit_model( X, NSD )


def GC_bias_correction( P, val ):
    """
    bins sequenece by GC content
    standardizes mean and variance for each bin
    """
    # bin NSD scores 
    MINBINSIZE = 5000
    # adapt bin size so that there are at least MINBINSIZE in each bin
    n = len(P.chrom)
    if n > MINBINSIZE:
        binsize = int( math.ceil( n/( float( math.floor( n/float(MINBINSIZE) ) ) ) ) )
    else:
        binsize = n

    P.CountBases()
    basecont = P.basecontent
    CGfrac,CpGratio = P.CpG_ratio()

    bint = [ (elem[1]+elem[2],idx) for idx,elem in enumerate(basecont) ]
    bint = sorted(bint)
    bin  = []
 
    for i,elem in enumerate(bint):
        try:
            bin[i/binsize] += [val[elem[1]]]
        except:
            bin.append( [ val[elem[1]] ] )

    mubin = []
    sdbin = []

    for i,elem in enumerate(bin):
        mubin += [ numpy.mean(bin[i]) ]
        sdbin += [ numpy.std(bin[i])  ]
        print 'GC bin: %d %4.2f %4.2f' % ( i, mubin[i], numpy.std(bin[i]) )

    for i,elem in enumerate(bint):
        val[elem[1]] -= mubin[i/binsize]
        val[elem[1]] /= sdbin[i/binsize]

    return val 



