import motifIO
import motifIO.base
import os,os.path,unittest
import motifanalysis
import motifanalysis.nsd
import random
import numpy,math
import seq.count as cnt
import seq._seq
import cistrome.regions as cis

def revcomp( l ):
    """
    Generate the reverse complement of a sequence.
    """
    comp = { 'A':'T', 'C':'G', 'G':'C', 'T':'A', 'N':'N' }
    l.reverse()
    r = [comp[x] for x in l]
    return r

def simulation(pssm,n=1000,w=200,mu0=0,sd0=1,mu1=1,sd1=1,f=0.05,sdpos=1,seed=1):
    """
    simulate DNA sequence and NSD scores
    pssm: motif matrix
    n: number of sequences to generare
    w: length of each sequence
    mu0: mean of score associated with unbound sequence
    sd0: sd of score associated with unbound sequence
    mu1: mean of score associated with bound sequence
    sd1: sd of score associated with bound sequence
    f: fraction of bound sequences
    sdpos: sd of binding around center of sequence
    """

    random.seed()

    base = ['A','C','G','T']
    motifseq = []
    for elem in pssm:
        row = []
        for i in range(4):
            row += elem[i]*[base[i]]
        motifseq.append(row)

    seql   = []
    scorel = []
    bindl  = []
    # generate random sequence 
    # for f percent of sequences generate a dna sequence from pssm and  score from N(mu1,sd1)
    mlen = len(motifseq)

    posl = []

    for i in range(n):
        seq  = [random.sample(base,1)[0] for i in range(w)]
        if random.random() < f:
            motif = [random.sample( motifseq[k],1)[0] for k in range(mlen) ]
            if random.random() < 0.5:
                motif = revcomp(motif)
            #pos = random.randint(0,w-5)
            pos = int(w*random.gauss(0.5,sdpos))
            if pos > 0 and pos < w-mlen-1:
                #pos = max(0,pos)
                #pos = min(w-mlen-1,pos)
                #print pos/float(w)
                posl += [pos+mlen/2.0]
                seq[pos:pos+mlen] = motif
                #print 'ttt',posl[-1],motif
            #print motif
            scorel += [random.gauss(mu1,sd1)]
            bindl += [1]
        else:
            scorel += [random.gauss(mu0,sd0)]
            bindl += [0]
        seql += [''.join(seq)]

    return seql,numpy.array(scorel),bindl



class TestNSD(unittest.TestCase):

    def test5(self):
        genome = "hg18"
        known_motifs = motifIO.base.MotifList()
        known_motifs_file_name = os.path.join(motifIO.PWMPATH,"transfac.xml")
        known_motifs.from_xml_file(known_motifs_file_name)
        test = 'M00959'  #ER
        datafile = os.path.join(os.path.split( __file__ )[0], 'data/ESR_MCF7_T45_FDR1.1000.bed')

        A = cis.interval('hg18','Estrogen receptor ChIP-chip in MCF-7')
        A.read_bed(datafile)
        A.TrimBed(length=500)
        A.getSequence(repeatMasked=False)
 
        for elem in known_motifs:
            if elem.id == test:
                pssm = numpy.array(elem.pssm,numpy.float)

        prob = cnt.count(A.seq)
        option = seq._seq.CUTOFF_OPTION
        #option = seq._seq.MAX_OPTION
        order  = 2

        idx,start,end,orient,score = seq._seq.seqscan( A.seq, pssm, prob, order, option )

        lengths = []

        for i,elem in enumerate(idx):
            print A.chrom[elem], A.start[elem], A.end[elem], start[i], end[i], score[i], A.seq[elem][ start[i]: end[i] ]
            lengths.append( len(A.seq[elem] ))

        results = motifanalysis.nsd.seqpos_stat( numpy.array(lengths), start, end, score, len(pssm) )
        self.assertTrue( results['pvalue'] < 1e-3 )


#    def test_NSD_select(self):
#        field = ["dist"]
#        low = [2000]
#        up  = [10000000]
#        phastcon = 0
#       nuc = ["LNCaP_H3K4me2_Vehicle_2Lanes_normalized_sorted.bed_nuc","L_D_H3K4me2_3Lanes_normalized_sorted.bed_nuc"]
#       lin = ["LNCaP_H3K4me2_Vehicle_2Lanes_normalized_sorted.bed_lin","L_D_H3K4me2_3Lanes_normalized_sorted.bed_lin"]
#       genome = "hg18"
#       datafile = os.path.join( os.path.split( __file__ )[0], 'data/LNCaP_H3K4me2_V_4_3h.pair.10.xls' )
#       P = motifanalysis.nsd.parse_table( datafile, genome=genome, select_id=['dist'], select_L=[2000], select_U=[9e9], nuc_id=nuc, lin_id=lin )
#       idx,treat_name = 0,nuc[0]
#      X = motifanalysis.nsd.NSD_select(P,idx,'xls',5,printnsd=True)
#      #test_nsd = [ 0.04157524, 0.03504730, 0.01742278, -0.01294448, -0.03426737, -0.03622983, -0.02095002, -0.04463528, 0.07545993, -0.02047828 ]
#      test_nsd = [0.04157524,0.03504730,0.01742278,-0.01294448,0.07545993]
#
#        for i,elem in enumerate(X.val):
#           self.assertTrue( abs( elem-test_nsd[i]) < 1e-5 )
#
#   def test_NSD_seqpos(self):
#       field = ["dist"]
#       low = [2000]
#       up  = [10000000]
#       phastcon = 0
#
#       nuc = ["veh_nuc","DHT_3h_nuc"]
#       lin = ["veh_lin","DHT_3h_lin"]
#       genome = "hg18"
#
#       known_motifs = motifIO.base.MotifList()
#        known_motifs_file_name = os.path.join(motifIO.PWMPATH,"transfac.xml")
#       known_motifs.from_xml_file(known_motifs_file_name)
#       test = ['M00447']
#
#       datafile = os.path.join( os.path.split( __file__ )[0], 'data/test.xls' )
#       out = motifanalysis.nsd.main_seqpos( datafile, known_motifs, names=test, genome=genome, select_id = ['dist'], select_U = [9e9], select_L = [2000], nuc_id=nuc, lin_id=lin, score_option='max', ntop=500, output_file='binoch_test_AR_position.txt', printnsd=False )
#       print out["DHT_3h_nuc"][0] 
#       self.assertEqual(1,1)
#       return
#
#
#   def test_bed_position(self):
#       genome = "hg18"
#       known_motifs = motifIO.base.MotifList()
#       known_motifs_file_name = os.path.join(motifIO.PWMPATH,"transfac.xml")
#       known_motifs.from_xml_file(known_motifs_file_name)
#       test = ['M00959']  #ER
#       datafile = os.path.join(os.path.split( __file__ )[0], 'data/ESR_MCF7_T45_FDR1.1000.bed' )
#       out = motifanalysis.nsd.main_seqpos( datafile, known_motifs, names=test, genome='hg18', span=600, score_option='max', ntop=0.01, output_file=None )
#       self.assertTrue( (out[-1][-1] < 1e-2) )
#
#
#   def test_sim_position(self):
#       for k in range(10):
#           pssm    = [[7,1,1,1],[1,7,1,1],[1,1,7,1],[1,1,1,7],[7,1,1,1],[1,7,1,1],[1,1,7,1],[1,1,1,7]]
#           n_set   = 1000
#           w_set   = 200
#           mu0_set = 0
#           sd0_set = 1
#           mu1_set = 5
#           sd1_set = 1
#
#           if k > 4:
#               f_set = 0.0
#               seql,NSD,bind = simulation(pssm,n=n_set,w=w_set,mu0=mu0_set,sd0=sd0_set,mu1=mu1_set,sd1=sd1_set,f=f_set,sdpos=100,seed=1)
#           else: 
#               f_set = 1.0
#               seql,NSD,bind = simulation(pssm,n=n_set,w=w_set,mu0=mu0_set,sd0=sd0_set,mu1=mu1_set,sd1=sd1_set,f=f_set,sdpos=0.1,seed=1)
#
#           prob = cnt.count(seql)
#
#           for elem in seql:
#               if len(elem) != w_set:
#                   sys.exit()
#
#           pssm = numpy.array(pssm,float)
#           for elem in pssm:
#               elem /= float(elem.sum())
#
#           s_idx,start,end,orient,score = seq._seq.seqscan( seql, pssm, prob, 2, motifanalysis.nsd.MAXPROB )
#           lengths = []
#
#           for i,elem in enumerate(s_idx):
#               lengths.append( len(seql[s_idx[elem]]) )
#           results = motifanalysis.nsd.seqpos_stat( numpy.array(lengths), start, end, score, len(pssm) )
#           if k < 5:
#               self.assertTrue( results['pvalue'] < 1e-3 )
#           else:
#                self.assertFalse( results['pvalue'] < 1e-3 )

if __name__ == "__main__":
    unittest.main()

