# Time-stamp: <2009-03-04 11:56:03 Tao Liu>

"""Module for PeakI, WigTrackI, FWTrackI and TrackI classes

Copyright (c) 2007,2008 Tao Liu <taoliu@jimmy.harvard.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the Artistic License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: taoliu@jimmy.harvard.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
import os
import re
import sys
import time
from   array import array
# ------------------------------------
# constants
# ------------------------------------
__version__ = "TabIO $Revision$"
__author__ = "Tao Liu <taoliu@jimmy.harvard.edu>"
__doc__ = "PeakIO, WigTrackI, FWTrackI and TrackI classes"

# ------------------------------------
# Misc functions
# ------------------------------------

# to determine the byte size
if array('H',[1]).itemsize == 2:
    BYTE2 = 'H'
else:
    raise Exception("BYTE2 type cannot be determined!")

if array('I',[1]).itemsize == 4:
    BYTE4 = 'I'
elif array('L',[1]).itemsize == 4:
    BYTE4 = 'L'
else:
    raise Exception("BYTE4 type cannot be determined!")

if array('f',[1]).itemsize == 4:
    FBYTE4 = 'f'
elif array('d',[1]).itemsize == 4:
    FBYTE4 = 'd'
else:
    raise Exception("BYTE4 type cannot be determined!")

# ------------------------------------
# Classes
# ------------------------------------

class PeakIO:
    """IO for peak region information.

    This class can hold peak information from MAT/MA2C/MACS, and
    provide some extra functions:

    1. filtering functions, filter_pvalue/score/fdr/fold are functions
    to filter the peaks according to pvalue/score/fdr/fold range given
    by the user.

    2. overlapping functions

    """
    def __init__ (self, comment=""):
        """Initialization function.

        comment: you can add any comments to the peakIO object like
        whether or not it is from a ChIP-chip or a ChIP-seq
        experiments.
        """
        self.peaks = {}
        self.comment = comment

    def dup (self):
        """return a duplicate peakI.
        """
        r = PeakIO(comment=self.comment)
        peaks = self.peaks
        new_peaks = {}
        chrs = peaks.keys()
        chrs.sort()
        for chrom in chrs:
            new_peaks[chrom]=[]
            new_peaks[chrom].extend(peaks[chrom])
        r.peaks = new_peaks
        return r
        
    
    def add (self, chromosome, start, end, summit=None, 
             score=None, total_p=None, 
             pvalue=None, fold_enrichment=None, fdr=None):
        """Use this function to add items to PeakIO object.

        items: (peak start,peak end, peak length, peak summit, peak
        score, number of tags/probes in peak region, peak pvalue, peak
        fold_enrichment, fdr) <-- tuple type

        Parameters:
        1. chromosome
        2. start
        3. end
        4. summit: the highest position for the peak region
        5. score:  the score for peak region
        6. total_p: total points in peak region. For ChIP-seq, it's
        how many tags in the region; for ChIP-chip, it's the number
        of probes.
        7. pvalue: -10*log(10,p-value) for peak region
        8. fold_enrichment: fold enrichment for the region
        9. fdr: False Discovery Rate for the region
        """
        if not self.peaks.has_key(chromosome):
            self.peaks[chromosome]=[]
        self.peaks[chromosome].append((start,end,end-start,summit,
                                       score,total_p,
                                       pvalue,fold_enrichment,fdr))
       
        
    def init_from_dict (self, data):
        """Initialize the data from a dictionary. Improper assignment
        will damage the data structure.
        
        """
        self.peaks = {}
        chrs = data.keys()
        chrs.sort()
        for chrom in chrs:
            self.peaks[chrom]=[]
            a = self.peaks[chrom].append
            for i in data[chrom]:
                a(i)

    def sort (self):
        peaks = self.peaks
        chrs = peaks.keys()
        chrs.sort()
        for chrom in chrs:
            peaks[chrom].sort(lambda x,y:cmp(x[0],y[0]))


    def get_chr_names (self):
        """Return all the chromosome names stored.
        
        """
        l = set(self.peaks.keys())
        return l

    def add_other_peakI (self,peak2):
        """Add another peakI object to self. 
        
        """
        peaks1 = self.peaks
        peaks2 = peak2.peaks
        chrs1 = peaks1.keys()
        chrs1.extend(peaks2.keys())
        chrs = set(chrs1)
        for chrom in chrs:
            if not peaks1.has_key(chrom):
                peaks1[chrom]=[]
            if peaks2.has_key(chrom):
                peaks1[chrom].extend(peaks2[chrom])
        
    def merge_overlap ( self ):
        """peak_candidates[chrom] = [(peak_start,peak_end,peak_length,peak_summit,peak_height,number_cpr_tags)...]

        """
        peaks = self.peaks
        new_peaks = {}
        chrs = peaks.keys()
        chrs.sort()
        for chrom in chrs:
            new_peaks[chrom]=[]
            n_append = new_peaks[chrom].append
            prev_peak = None
            peaks_chr = peaks[chrom]
            for i in xrange(len(peaks_chr)):
                if not prev_peak:
                    prev_peak = peaks_chr[i]
                    continue
                else:
                    if peaks_chr[i][0] <= prev_peak[1]:
                        s_new_peak = prev_peak[0]
                        e_new_peak = max(peaks_chr[i][1],prev_peak[1])
                        l_new_peak = e_new_peak-s_new_peak
                        if peaks_chr[i][4] > prev_peak[4]:
                            summit_new_peak = peaks_chr[i][3]
                            h_new_peak = peaks_chr[i][4]
                        else:
                            summit_new_peak = prev_peak[3]
                            h_new_peak = prev_peak[4]
                        prev_peak = (s_new_peak,e_new_peak,l_new_peak,summit_new_peak,h_new_peak,peaks_chr[i][5]+prev_peak[5])
                    else:
                        n_append(prev_peak)
                        prev_peak = peaks_chr[i]
            if prev_peak:
                n_append(prev_peak)
        del peaks
        self.peaks = new_peaks
        return True

    def overlap_with_other_peaks (self, peaks2, cover=0):
        """Peaks2 is a PeakIO object or dictionary with can be
        initialzed as a PeakIO. check __init__ for PeakIO for detail.

        return how many peaks are intersected by peaks2 by percentage
        coverage on peaks2(if 50%, cover = 0.5).
        """
        peaks1 = self.peaks
        if isinstance(peaks2,PeakIO):
            peaks2 = peaks2.peaks
        total_num = 0
        chrs1 = peaks1.keys()
        chrs2 = peaks2.keys()
        for k in chrs1:
            if not chrs2.count(k):
                continue
            rl1_k = iter(peaks1[k])
            rl2_k = iter(peaks2[k])
            tmp_n = False
            try:
                r1 = rl1_k.next()
                r2 = rl2_k.next()
                while (True):
                    if r2[0] < r1[1] and r1[0] < r2[1]:
                        a = sorted([r1[0],r1[1],r2[0],r2[1]])
                        if float(a[2]-a[1]+1)/r2[2] > cover:
                            if not tmp_n:
                                total_num+=1
                                tmp_n = True
                    if r1[1] < r2[1]:
                        r1 = rl1_k.next()
                        tmp_n = False
                    else:
                        r2 = rl2_k.next()
            except StopIteration:
                continue
        return total_num

    def overlap_with_FWTrackI (self, fwt, cover=0, twostrand=True):
        """
        
        """
        self_chrs = list(self.get_chr_names())
        if twostrand == False:
            fwt.self_merge_plus_minus_ranges_w_duplicates() # merge + and - strand
        fwt_chrs = list(fwt.get_chr_names())
        w = fwt.fw
        v = {}
        total_num = 0
        for k in self_chrs:
            if not fwt_chrs.count(k):
                continue
            # + strand fwt
            peak_k = iter(self.peaks[k])
            fwt_k = iter(fwt.get_ranges_by_chr(k)[0])
            tmp_n = False
            try:
                p1 = peak_k.next()
                f2b = fwt_k.next()
                f2e = f2b+w
                pt = k+":"+str(p1[0])+'...'+str(p1[1])
                tt = ""
                while (True):

                    if f2b < p1[1] and p1[0] < f2e:
                        a = sorted([p1[0],p1[1],f2b,f2e])
                        if float(a[2]-a[1]+1)/w > cover:
                            if not tmp_n:
                                total_num+=1
                                tmp_n = True
                        
                        tt += "\t"+str(f2b)+'...'+str(f2e)+"+"
                    if p1[1] < f2e:
                        if tt.find('\t') != -1:
                            v[pt]=tt
                        #print t
                        p1 = peak_k.next()
                        pt = k+":"+str(p1[0])+'...'+str(p1[1])
                        tt = ""
                        tmp_n = False
                    else:
                        f2b = fwt_k.next()
                        f2e = f2b + w
            except StopIteration:
                if tt and tt.find('\t') != -1:
                    v[pt]=tt
            # - strand fwt
            peak_k = iter(self.peaks[k])
            fwt_k = iter(fwt.get_ranges_by_chr(k)[1])
            tmp_n = False
            try:
                p1 = peak_k.next()
                f2e = fwt_k.next()
                f2b = f2e-w

                pt = k+":"+str(p1[0])+'...'+str(p1[1])
                tt = ""
                while (True):
                    if f2b < p1[1] and p1[0] < f2e:
                        a = sorted([p1[0],p1[1],f2b,f2e])
                        if float(a[2]-a[1]+1)/w > cover:
                            if not tmp_n:
                                total_num+=1
                                tmp_n = True
                
                        tt += "\t"+str(f2b)+'...'+str(f2e)+"-"
                    if p1[1] < f2e:
                        if tt.find('\t') != -1:
                            if v.has_key(pt):
                                v[pt]+=tt
                            else:
                                v[pt]=tt
                        #print t
                        p1 = peak_k.next()
                        pt = k+":"+str(p1[0])+'...'+str(p1[1])
                        tt = ""
                        tmp_n = False
                    else:
                        f2e = fwt_k.next()
                        f2b = f2e - w
            except StopIteration:
                if tt and tt.find('\t') != -1:
                    if v.has_key(pt):
                        v[pt]+=tt
                    else:
                        v[pt]=tt

                continue

#--------------------------------------------------------
#  Fill in zero counts 
#--------------------------------------------------------
        for k in self_chrs:
            if not fwt_chrs.count(k):
                continue
            peak_k = iter(self.peaks[k])
            try:
                while (True):
                    p1 = peak_k.next()
                    pt = k+":"+str(p1[0])+'...'+str(p1[1])
                    if not v.has_key(pt):
                        v[pt]=""
            except StopIteration:
                continue
#--------------------------------------------------------

        return (len(v.keys()),v)
        
