# names of genes associated with nearest sites

import sys
import numpy
import cistrome.regions as cb
import table.BedIO      as BedIO
import table.TagIO      as TagIO
import table.PeakIO     as PeakIO
import math
import optparse

def peak_pair( bed, minsep=250, maxsep=450 ):
    """
    Find pairs of peaks with midpoint to midpoint distance in\n
    range: [minsep,maxsep).
    """

    peaks = PeakIO.PeakIO()
    idx = 0

    for chr in bed.peaks:
        l = bed.peaks[chr]
        l.sort()

        # first pair on list
        start,end,mid    = l[1][0],l[1][1],int(0.5*(l[1][0]+l[1][1]))
        pstart,pend,pmid = l[0][0],l[0][1],int(0.5*(l[0][0]+l[0][1]))
        delta = mid - pmid

        # if criterion is met enter interval with extremes of intervals
        if delta >= minsep and delta < maxsep:
            peaks.add(chr,pstart,end,pend,start,1,1)

        for i,elem in enumerate(l[2:]):
            start,end,mid = elem[0],elem[1],int(0.5*(elem[0]+elem[1]))
   
            # test nearest two nucleosomes for correct position
            for k in [1,0]:
                pstart,pend = l[i+k][0],l[i+k][1]
                pmid = int(0.5*(pstart+pend) )
                delta = mid - pmid

                # if criterion is met enter interval with extremes of intervals
                if delta >= minsep and delta < maxsep:
                    peaks.add(chr,pstart,end,pend,start,1,1)
                    #print delta

    return peaks
#return self.peaks[chromosome].append((start,end,end-start,summit,score,total_p,pvalue,fold_enrichment,fdr))

def peak_trim( bed, width=200 ):
    """
    Trim intervals to be of specified width around the original midpoint
    """
    delta = int(width/2.0)
    new = PeakIO.PeakIO()
    for chr in bed.peaks:
        l = bed.peaks[chr]
        l.sort()
        for i,elem in enumerate(l):
            start,end = elem[0],elem[1]
            #print chr,start,end
            mid = int(0.5*(start+end))
            new.add(chr,mid-delta,mid+delta,1,1,1,1)
    return new

def peak_extend( bed, shift ):
    """
    Extend both ends of every interval by shift bp
    """
    new = PeakIO.PeakIO()
    for chr in bed.peaks:
        l = bed.peaks[chr]
        for i,elem in enumerate(l):
            start,end = elem[0],elem[1]
            new.add(chr,start-shift,end+shift,1,1,1,1)
    return new


def peak_split(bed,n=4):
    """
    separates odd and even index peaks into n objects
    in this use ( but not in generat ) ensures intervals dont overlap
    """

    newl = [ PeakIO.PeakIO() for i in range(n) ]
    #new0 = PeakIO.PeakIO()
    #new1 = PeakIO.PeakIO()
    for chr in bed.peaks:
        l = bed.peaks[chr]
        for i,elem in enumerate(l):
            start,end = elem[0],elem[1]
            newl[i%n].add(chr,start,end,1,1,1,1)
            #if i % 2 == 0:
            #    new0.add(chr,start,end,1,1,1,1)
            #else:
            #    new1.add(chr,start,end,1,1,1,1)
    #return (new0,new1)
    return newl

def key_to_pos(key):
    """
    converts string of format chr*:1020....1120
    to tuple ( chr*, 1020, 1120 )
    """
    f = key.split(':')
    chr = f[0]
    f = f[1].split( '...' )
    start, end = int(f[0]), int(f[1])
    return (chr,start,end)

def tagstr_to_list(tagstr):
    a = []
    tags = tagstr.split('\t')
    for z in tags[1:]:
        strand = z[-1]
        t = z[:-1].split( '...' )
        a += [(int(t[0]),int(t[1]),strand)]
    return a

def count_tags_in_single_nuc(a,start,end,taglen=25,nuclen=200,shift=73,remove_redundancy=False):
    """
    assumes overlap is done on unshifted tags with shifted starts and end - and must be accounted for
    note elem[0] of a is the 5' end of the tag 
    """
    nuc = 0
    d = {}
    for elem in a:

        strand = elem[2]

        if strand == '+':
            pt_s = elem[0] + shift
            pt_e = elem[0] + taglen + shift
        else:
            pt_s = elem[0] - taglen - shift
            pt_e = elem[0] - shift

        if remove_redundancy == False or (pt_s,strand) not in d:
            if ( pt_s >= start+shift and pt_e < end-shift ):
                nuc += 1
            #elif ( pt >= start + shift + nuclen and pt < end - shift - nuclen ):
            #    linker += 1
            d[(pt_s,strand)] = 1

    return nuc

def count_tags_in_nuc(a,start,end,taglen=25,nuclen=200,shift=73,remove_redundancy=False):
    """
    assumes overlap is done on unshifted tags with shifted starts and end - and must be accounted for
    we want overlap shifted tags with unshifted starts and ends
    note elem[0] of a is the 5' end of the tag 
    """
    nuc,linker = 0,0
    d = {}

    for elem in a:
        #print "**", elem
        strand = elem[2]

        # shift tags
        if strand == '+':
            pt_s = elem[0] + shift
            pt_e = elem[0] + shift + taglen
        else:
            pt_s = elem[0] - shift - taglen
            pt_e = elem[0] - shift

        #print start,end,elem[0],strand

        if remove_redundancy == False or (pt_s,strand) not in d:
            if ( pt_e < start + shift + nuclen and pt_s >= start + shift ) or ( pt_s >= end - shift - nuclen and pt_e < end - shift ):
                nuc += 1
            elif ( pt_s >= start + shift + nuclen and pt_e < end - shift - nuclen ):
                linker += 1
            d[(pt_s,strand)] = 1

    return (nuc,linker)


def nsites( bed ):
    """
    count the number of sites in TL's bed object
    """
    n = 0
    for elem in bed.peaks:
        n += len( bed.peaks[elem] )
    return n


def extract_tags( genome, nucleosome_bed, tag_bed_list, chip_list, minsep=250, maxsep=450, nuclen=200, shift=73, taglength=25, remove_redundancy=False, interval_type="pair", output=None ):

    """
    Pair nucleosomes in close proximity
    For interval_type = paired
    Count the number of (3' shifted tags) in nucleosome and in linker region
    """

    # parse nucleosome file
    peakfhd   = file(nucleosome_bed)
    peaktrack = BedIO.parse_BED(peakfhd)
    peakt     = peak_trim(peaktrack,width=nuclen)

    if interval_type == "pair":
        pair = peak_pair(peakt, minsep=minsep, maxsep=maxsep)
    elif interval_type == "mono":
        pair = peakt

    pair  = peak_extend( pair, shift )
    pairl = peak_split(pair)

    if output is None:
        fp = sys.stdout
    else:
        fp = open( output, 'w' )
        print >> fp, "# number of regions: %d" %nsites(pair)

    # parse tag files
    tag = {}
    for elem in tag_bed_list:
        fh = file(elem)
        tag[elem] = TagIO.BEDParser().build_fwtrack(fh)
        tag[elem].fw = taglength
        fh.close() 

    new = cb.interval(genome)
    new.description = ''
    new.chrom,new.start,new.end,new.info,new.strand,new.mid,new.length,new.name  = [],[],[],[],[],[],[],[]

    n_overlap,v_overlap = {},{}
    keylist = {}

    #for pair in [pair0,pair1]:
    for pair in pairl:

        for elem in tag_bed_list:
            (n_overlap[elem],v_overlap[elem]) = pair.overlap_with_FWTrackI(tag[elem])

        allkeys = []
        for elem in v_overlap.keys():
            allkeys += v_overlap[elem].keys()

        #for key in v_overlap[ v_overlap.keys()[0] ].keys() + v_overlap[ v_overlap.keys()[1] ].keys():
        for key in allkeys:

            if key not in keylist:

                # here ends are shifted back to original positions !
                chr,start,end = key_to_pos( key )
 
                new.chrom  += [chr]
                new.start  += [start+shift]
                new.mid    += [int(0.5*(start+end))]
                new.length += [end-start-2*shift]
                new.end    += [end-shift]
                new.strand += ['+']
                new.name   += [key]
        
                info = []

                if interval_type == "pair":

                    for elem in tag_bed_list:
                        if key in v_overlap[elem]:
                            a = tagstr_to_list( v_overlap[elem][key] )
                            nuc,linker = count_tags_in_nuc(a,start,end,taglen=taglength,nuclen=nuclen,shift=shift,remove_redundancy=remove_redundancy)
                        else:
                            nuc,linker = 0,0
                        info += [nuc,linker]

                elif interval_type == "mono":

                    for elem in tag_bed_list:
                        if key in v_overlap[elem]:
                            a = tagstr_to_list( v_overlap[elem][key] )
                            nuc = count_tags_in_single_nuc(a,start,end,taglen=taglength,nuclen=nuclen,shift=shift,remove_redundancy=remove_redundancy)
                        else:
                            nuc = 0
                        info += [nuc]

                new.info += [ [str(x) for x in info ] ]
                keylist[key] = True

    new = new.order()
    new.name = [ 'ID%d' % k for k,x in enumerate(new.chrom) ]

    dist = []
    ID   = []

    if chip_list:
        for chip in chip_list:
            if chip != None:
                R = cb.interval(genome,description=chip)
                R.read_bed(chip) 
                R.strand = [ '+' for x in R.chrom ]
                R.info   = [ ''  for x in R.chrom ]
                if not R.name:
                    R.name   = [ 'ID%d' % k for k,x in enumerate(R.chrom) ]
     
                new.distance(R,nsites=1,distance_type='abs')
                dist.append([])
                ID.append([])
                for k,x in enumerate(new.nearest_site):
                    if type(x) == type([]) and len(x) > 0:
                        try:
                            dist[-1].append( x[0] )
                            ID[-1].append( new.nearest_site_name[k][0] )
                        except:
                            pass
                    else:
                        dist[-1].append( cb.NONEARESTSITE )
                        ID[-1].append( "" )

    print >>fp,"# genome: %s" % genome
    print >>fp,"# nucleosome position file: %s" % nucleosome_bed
    print >>fp,"# ChIP-seq tag files: %s" % "\t".join(tag_bed_list)
    
    if chip_list:
        print >>fp,"# annotation bed files: %s" % "\t".join(chip_list)
    
    if interval_type == "pair":
        print >>fp,"# minsep: %d" % minsep
        print >>fp,"# maxsep: %d" % maxsep
    
    print >>fp,"# shift: %d" % shift
    print >>fp,"# tag length: %d" % taglength
    print >>fp,"# remove redundancy:", remove_redundancy
    print >>fp,"# interval type: %s" % interval_type

    print >> fp, "chrom\tstart\tend\tid\t",
    if chip_list:
        print >> fp, "\t".join( [ "%s.dist" % elemx for elemx in  chip_list ] ),
        print >> fp, "\t".join( [ "%s.ID"   % elemx for elemx in  chip_list ] ),

    if interval_type == "mono": 
        print >> fp, "\t" + "\t".join(tag_bed_list)
    elif interval_type == "pair": 
        print >> fp, "\t" + "\t".join( [ elem+'_nuc\t'+elem+'_lin' for elem in tag_bed_list ] )

    for i,elem in enumerate( new.chrom ):
        try:
            print >> fp, "%s\t%d\t%d\t%s" % ( new.chrom[i], new.start[i], new.end[i], new.name[i] ),
            if dist:
                for k in range( len(dist) ):
                    print >> fp, "%d\t" % ( int( dist[k][i] ) ),
                for k in range( len(dist) ):
                    print >> fp, "%s\t" % ( ID[k][i] ),
            print >> fp, "\t".join( new.info[i] )
        except:
            print new.chrom[i]#,dist[k][i]

    return new


def check_files(fl):
    for elem in fl:
        try:
            fh = open( elem )
            fh.close()
        except:
            print "file not found: %s" % elem
            return False
    return True


class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg


#def main( genome, regionfile, tag, annot, minsep=250, maxsep=450, nuclen=200, shift=73, taglength=25, remove_redundancy=False, interval_type="pair", output=None ):
def main(args=None):

    sep = ','

    if args is None:
        args = sys.argv

    try:
        try:
            usage = """Usage: tagtab [options] region_file tag_files ( separated by commas ) """
            parser = optparse.OptionParser(usage=usage)
            parser.add_option( '-a', '--annot', dest='annot', default=None, help="comma delimited list of annotation bed files. The distance to the nearest regions in each of these files appears a column in the table. Bedfiles from UCSC tables can be downloaded from: http://genome.ucsc.edu/cgi-bin/hgTables?command=start" )
            parser.add_option( '-g', '--genome',    dest='genome',     default='hg18' )
            parser.add_option( '-m', '--model',     dest='model',      default='pair', help="mono or pair, mono outputs only counts in region, pair outputs data for NSD analysis" )
            parser.add_option( '--minsep',          dest='minsep',     default=250,   help="Minimum center to center distance for nucleosome pairs.  (must be > nuclen and < maxsep) " )
            parser.add_option( '--maxsep',          dest='maxsep',     default=450,   help="Maximum center to center distance for nucleosome pairs.  (must be > minsep and < 3 x nuclen )" )
            parser.add_option( '-l','--nuclen',     dest='nuclen',     default=200,   help="Span of nucleosome region." )
            parser.add_option( '-s','--shift',      dest='shift',      default=73,    help="Amount by which tags are shifted 3' before counting" )
            parser.add_option( '-t','--taglength',  dest='taglength',  default=25,    help="Length of ChIP-seq tag sequences." )
            parser.add_option( '-r','--redundancy', dest='redundancy', default=False, help="Set to True if redundant tags are to be removed, False otherwise." )
            parser.add_option( '-o','--output',     dest='output',     default=None,  help="Name of file to which table is to be written." )

            (opts,args) = parser.parse_args(sys.argv)

            regionfile = args[1]
 
            if opts.annot is not None:
                if sep in opts.annot:
                    annotfile = opts.annot.split(sep)
                else:
                    annotfile = [opts.annot]
            else:
                annotfile = None

            if annotfile:
                for elem in annotfile:
                    try:
                        open(elem,'r')
                    except IOError:
                        raise Usage( "Your annotation file %s cannot be opened.\n\n\n%s" % (elem,usage) )

            if opts.model not in ('mono','pair'):
                raise Usage("Model must be specified as \'mono\' or \'pair\'.")

            if opts.model == 'pair':
                if int(opts.nuclen) < 100 or int(opts.nuclen) > 300:
                    raise Usage("nuclen must be in range [100,300].")
                if int(opts.minsep) < int(opts.nuclen):
                    raise Usage("minsep must be > nuclen.")
                if int(opts.maxsep) <= int(opts.minsep):
                    raise Usage("minsep must be < maxsep.")
                if int(opts.maxsep) > 3*int(opts.nuclen):
                    raise Usage("maxsep must be < 3 x nuclen.")

        except IndexError or parser.error, usage:
            raise Usage(usage)
            return 2

        try:
            open(regionfile,'r')
        except (IOError,IndexError):
            raise Usage( "Your input file \'%s\' cannot be opened.\n\n\n%s" % (regionfile,usage) )
            return 2

        tag = args[2].split(sep)
        for elem in tag:
            try:
                open(elem,'r')
            except (IOError,IndexError):
                raise Usage( "Your input file \'%s\' cannot be opened.\n\n\n%s" % (elem,usage) )
                return 2

        extract_tags( opts.genome, regionfile, tag, annotfile, minsep=int(opts.minsep), maxsep=int(opts.maxsep), nuclen=int(opts.nuclen), shift=int(opts.shift), taglength=int(opts.taglength), remove_redundancy=opts.redundancy, interval_type=opts.model, output=opts.output )

    except Usage,err:
        print >>sys.stderr, err.msg
        print >>sys.stderr, "for help use --help"
        return 2


if __name__ == '__main__':
    sys.exit(main())
